/*
 * lzx-comp.c
 *
 * LZX compression routines. This is based on code written by Matthew T.
 * Russotto.
 *
   Copyright (C) 2002 Matthew T. Russotto
 * Copyright (C) 2012 Eric Biggers
 *
 * wimlib - Library for working with WIM files 
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */


#include <string.h>
#include <math.h>

#include "util.h"
#include "lzx.h"
#include "endianness.h"

/* Intermediate (non-leaf) node in a Huffman tree. */
typedef struct HuffmanNode {
	u32 freq;
	u16 sym;
	union {
		u16 path_len;
		u16 height;
	};
	struct HuffmanNode *left_child;
	struct HuffmanNode *right_child;
} HuffmanNode;

/* Leaf node in a Huffman tree.  The fields are in the same order as the
 * HuffmanNode, so it can be cast to a HuffmanNode.  There are no pointers to
 * the children in the leaf node. */
typedef struct {
	u32 freq;
	u16 sym;
	union {
		u16 path_len;
		u16 height;
	};
} HuffmanLeafNode;

/* Structure to contain the Huffman codes for the main, length, and aligned
 * offset trees. */
struct lzx_codes {
	u16 main_codewords[MAINTREE_NUM_SYMBOLS];
	u8  main_lens[MAINTREE_NUM_SYMBOLS];

	u16 len_codewords[LENTREE_NUM_SYMBOLS];
	u8  len_lens[LENTREE_NUM_SYMBOLS];

	u16 aligned_codewords[ALIGNEDTREE_NUM_SYMBOLS];
	u8  aligned_lens[ALIGNEDTREE_NUM_SYMBOLS];
};

struct lzx_freq_tables {
	u32 main_freq_table[MAINTREE_NUM_SYMBOLS]; 
	u32 len_freq_table[LENTREE_NUM_SYMBOLS];
	u32 aligned_freq_table[ALIGNEDTREE_NUM_SYMBOLS];
};


/* Structure to keep track of the current position in the uncompressed input. */
struct input_byte_buffer {
	unsigned num_bytes_remaining;
	u8 *data;
};

typedef u16 bitbuf_t;

/* Structure to keep track of the current position in the compressed output. */
struct output_bit_buffer {
	/* A variable to buffer writing bits to the output and is flushed to the
	 * compressed output when full. */
	bitbuf_t bitbuf;

	/* Number of bits in @bitbuf that are valid. */
	unsigned filled_bits;

	/* Pointer to the next byte in the compressed output. */
	u8 *output;

	/* Number of space left in the memory pointed to by @output. */
	unsigned num_bytes_remaining;
};

/* Comparator function for HuffmanLeafNodes.  Sorts primarily by symbol
 * frequency and secondarily by symbol value. */
static int cmp_leaves_by_freq(const void *__leaf1, const void *__leaf2)
{
	const HuffmanLeafNode *leaf1 = __leaf1;
	const HuffmanLeafNode *leaf2 = __leaf2;

	int freq_diff = (int)leaf1->freq - (int)leaf2->freq;

	if (freq_diff == 0)
		return (int)leaf1->sym - (int)leaf2->sym;
	else
		return freq_diff;
}

/* Comparator function for HuffmanLeafNodes.  Sorts primarily by code length and
 * secondarily by symbol value. */
static int cmp_leaves_by_code_len(const void *__leaf1, const void *__leaf2)
{
	const HuffmanLeafNode *leaf1 = __leaf1;
	const HuffmanLeafNode *leaf2 = __leaf2;

	int code_len_diff = (int)leaf1->path_len - (int)leaf2->path_len;

	if (code_len_diff == 0)
		return (int)leaf1->sym - (int)leaf2->sym;
	else
		return code_len_diff;
}

/* Recursive function to calculate the depth of the leaves in a Huffman tree.
 * */
static void huffman_tree_compute_path_lengths(HuffmanNode *node, u16 cur_len)
{
	if (node->sym == (u16)(-1)) {
		/* Intermediate node. */
		huffman_tree_compute_path_lengths(node->left_child, cur_len + 1);
		huffman_tree_compute_path_lengths(node->right_child, cur_len + 1);
	} else {
		/* Leaf node. */
		node->path_len = cur_len;
	}
}

/* Creates a canonical Huffman code from an array of symbol frequencies. 
 *
 * The algorithm used is similar to the well-known algorithm that builds a
 * Huffman tree using a minheap.  In that algorithm, the leaf nodes are
 * initialized and inserted into the minheap with the frequency as the key.
 * Repeatedly, the top two nodes (nodes with the lowest frequency) are taken out
 * of the heap and made the children of a new node that has a frequency equal to
 * the sum of the two frequencies of its children.  This new node is inserted
 * into the heap.  When all the nodes have been removed from the heap, what
 * remains is the Huffman tree. The Huffman code for a symbol is given by the
 * path to it in the tree, where each left pointer is mapped to a 0 bit and each
 * right pointer is mapped to a 1 bit.
 *
 * The algorithm used here uses an optimization that removes the need to
 * actually use a heap.  The leaf nodes are first sorted by frequency, as
 * opposed to being made into a heap.  Note that this sorting step takes O(n log
 * n) time vs.  O(n) time for heapifying the array, where n is the number of
 * symbols. However, the heapless method is probably faster overall, due to the
 * time saved later.  In the heapless method, whenever an intermediate node is
 * created, it is not inserted into the sorted array.  Instead, the intermediate
 * nodes are kept in a separate array, which is easily kept sorted because every
 * time an intermediate node is initialized, it will have a frequency at least
 * as high as that of the previous intermediate node that was initialized.  So
 * whenever we want the 2 nodes, leaf or intermediate, that have the lowest
 * frequency, we check the low-frequency ends of both arrays, which is an O(1)
 * operation.
 *
 * The function builds a canonical Huffman code, not just any Huffman code.  A
 * Huffman code is canonical if the codeword for each symbol numerically
 * precedes the codeword for all other symbols of the same length that are
 * numbered higher than the symbol, and additionally, all shorter codewords,
 * 0-extended, numerically precede longer codewords.  A canonical Huffman code
 * is useful because it can be reconstructed by only knowing the path lengths in
 * the tree.  See the make_decode_table() function in lzx-decomp.c to see how to
 * reconstruct a canonical Huffman code from only the lengths of the codes.
 *
 * @num_symbols:  The number of symbols in the alphabet.
 *
 * @max_codeword_len:  The maximum allowed length of a codeword in the code.
 * 			Note that if the code being created runs up against
 * 			this restriction, the code ultimately created will be
 * 			suboptimal, although there are some advantages for
 * 			limiting the length of the codewords.
 *
 * @symbol_freqs:  An array of length @num_symbols that contains the frequencies
 * 			of each symbol in the uncompressed data.
 *
 * @lens:	   An array of length @num_symbols into which the lengths of the
 * 			codewords for each symbol will be written.
 *
 * @codewords:	   An array of @num_symbols short integers into which the
 * 			codewords for each symbol will be written.  The first 
 * 			lens[i] bits of codewords[i] will contain the codeword 
 * 			for symbol i.
 */
static void make_canonical_huffman_code(unsigned num_symbols, 
			unsigned max_codeword_len, const unsigned symbol_freqs[],
						u8 lens[], u16 codewords[])
{
	/* We require at least 2 possible symbols in the alphabet to produce a
	 * valid Huffman decoding table. It is allowed that fewer than 2 symbols
	 * are actually used, though. */
	wimlib_assert(num_symbols >= 2);

	/* Initialize the lengths and codewords to 0 */
	memset(lens, 0, num_symbols * sizeof(lens[0]));
	memset(codewords, 0, num_symbols * sizeof(codewords[0]));

	/* Calculate how many symbols have non-zero frequency.  These are the
	 * symbols that actually appeared in the input. */
	unsigned num_used_symbols = 0;
	for (unsigned i = 0; i < num_symbols; i++)
		if (symbol_freqs[i] != 0)
			num_used_symbols++;


	/* It is impossible to make a code for num_used_symbols symbols if there
	 * aren't enough code bits to uniquely represent all of them. */
	wimlib_assert((1 << max_codeword_len) > num_used_symbols);

	/* Initialize the array of leaf nodes with the symbols and their
	 * frequencies. */
	HuffmanLeafNode leaves[num_used_symbols];
	unsigned leaf_idx = 0;
	for (int i = 0; i < num_symbols; i++) {
		if (symbol_freqs[i] != 0) {
			leaves[leaf_idx].freq = symbol_freqs[i];
			leaves[leaf_idx].sym  = i;
			leaves[leaf_idx].height = 0;
			leaf_idx++;
		}
	}

	/* Deal with the special cases where num_used_symbols < 2. */
	if (num_used_symbols < 2) {
		if (num_used_symbols == 0) {
			/* If num_used_symbols is 0, there are no symbols in the
			 * input, so it must be empty.  This should be an error,
			 * but the LZX format expects this case to succeed.  All
			 * the codeword lengths are simply marked as 0 (which was
			 * already done.) */
		} else {
			/* If there only one symbol is present, the LZX format
			 * requires that the Huffman code include two codewords.
			 * One is not used.  Note that this doesn't make the
			 * encoded data take up more room anyway, since binary
			 * data itself has 2 symbols. */

			unsigned sym = leaves[0].sym;

			codewords[0] = 0;
			lens[0]      = 1;
			if (sym == 0) {
				/* dummy symbol is 1, real symbol is 0 */
				codewords[1] = 1;
				lens[1]      = 1;
			} else {
				/* dummy symbol is 0, real symbol is sym */
				codewords[sym] = 1;
				lens[sym]      = 1;
			}
		}
		return;
	}

	/* Otherwise, there are at least 2 symbols in the input, so we need to
	 * find a real Huffman code. */




	/* Declare the array of intermediate nodes.  An intermediate node is not
	 * associated with a symbol. Instead, it represents some binary code
	 * prefix that is shared between at least 2 codewords.  There can be at
	 * most num_used_symbols - 1 intermediate nodes when creating a Huffman
	 * code.  This is because if there were at least num_used_symbols nodes,
	 * the code would be suboptimal because there would be at least one
	 * unnecessary intermediate node.  
	 *
	 * The worst case (greatest number of intermediate nodes) would be if
	 * all the intermediate nodes were chained together.  This results in
	 * num_used_symbols - 1 intermediate nodes.  If num_used_symbols is at
	 * least 17, this configuration would not be allowed because the LZX
	 * format constrains codes to 16 bits or less each.  However, it is
	 * still possible for there to be more than 16 intermediate nodes, as
	 * long as no leaf has a depth of more than 16.
	 */
	HuffmanNode inodes[num_used_symbols - 1];


	/* Pointer to the leaf node of lowest frequency that hasn't already been
	 * added as the child of some intermediate note. */
	HuffmanLeafNode *cur_leaf = &leaves[0];

	/* Pointer past the end of the array of leaves. */
	HuffmanLeafNode *end_leaf = &leaves[num_used_symbols];

	/* Pointer to the intermediate node of lowest frequency. */
	HuffmanNode     *cur_inode = &inodes[0];

	/* Pointer to the next unallocated intermediate node. */
	HuffmanNode     *next_inode = &inodes[0];

	/* Only jump back to here if the maximum length of the codewords allowed
	 * by the LZX format (16 bits) is exceeded. */
try_building_tree_again:

	/* Sort the leaves from those that correspond to the least frequent
	 * symbol, to those that correspond to the most frequent symbol.  If two
	 * leaves have the same frequency, they are sorted by symbol. */
	qsort(leaves, num_used_symbols, sizeof(leaves[0]), cmp_leaves_by_freq);

	cur_leaf   = &leaves[0];
	cur_inode  = &inodes[0];
	next_inode = &inodes[0];

	/* The following loop takes the two lowest frequency nodes of those
	 * remaining and makes them the children of the next available
	 * intermediate node.  It continues until all the leaf nodes and
	 * intermediate nodes have been used up, or the maximum allowed length
	 * for the codewords is exceeded.  For the latter case, we must adjust
	 * the frequencies to be more equal and then execute this loop again. */
	while (1) {

		/* Lowest frequency node. */
		HuffmanNode *f1 = NULL; 

		/* Second lowest frequency node. */
		HuffmanNode *f2 = NULL;

		/* Get the lowest and second lowest frequency nodes from
		 * the remaining leaves or from the intermediate nodes.
		 * */

		if (cur_leaf != end_leaf && (cur_inode == next_inode || 
					cur_leaf->freq <= cur_inode->freq)) {
			f1 = (HuffmanNode*)cur_leaf++;
		} else if (cur_inode != next_inode) {
			f1 = cur_inode++;
		}

		if (cur_leaf != end_leaf && (cur_inode == next_inode || 
					cur_leaf->freq <= cur_inode->freq)) {
			f2 = (HuffmanNode*)cur_leaf++;
		} else if (cur_inode != next_inode) {
			f2 = cur_inode++;
		}

		/* All nodes used up! */
		if (f1 == NULL || f2 == NULL)
			break;

		/* next_inode becomes the parent of f1 and f2. */

		next_inode->freq   = f1->freq + f2->freq;
		next_inode->sym    = (u16)(-1); /* Invalid symbol. */
		next_inode->left_child   = f1;
		next_inode->right_child  = f2;

		/* We need to keep track of the height so that we can detect if
		 * the length of a codeword has execeed max_codeword_len.   The
		 * parent node has a height one higher than the maximum height
		 * of its children. */
		next_inode->height = max(f1->height, f2->height) + 1;

		/* Check to see if the code length of the leaf farthest away
		 * from next_inode has exceeded the maximum code length. */
		if (next_inode->height > max_codeword_len) {
			/* The code lengths can be made more uniform by making
			 * the frequencies more uniform.  Divide all the
			 * frequencies by 2, leaving 1 as the minimum frequency.
			 * If this keeps happening, the symbol frequencies will
			 * approach equality, which makes their Huffman
			 * codewords approach the length
			 * log_2(num_used_symbols).
			 * */
			for (unsigned i = 0; i < num_used_symbols; i++)
				if (leaves[i].freq > 1)
					leaves[i].freq >>= 1;
			goto try_building_tree_again;
		} 
		next_inode++;
	}

	/* The Huffman tree is now complete, and its height is no more than
	 * max_codeword_len.  */

	HuffmanNode *root = next_inode - 1;
	wimlib_assert(root->height <= max_codeword_len);

	/* Compute the path lengths for the leaf nodes. */
	huffman_tree_compute_path_lengths(root, 0);

	/* Sort the leaf nodes primarily by code length and secondarily by
	 * symbol.  */
	qsort(leaves, num_used_symbols, sizeof(leaves[0]), cmp_leaves_by_code_len);

	u16 cur_codeword = 0;
	unsigned cur_codeword_len = 0;
	for (unsigned i = 0; i < num_used_symbols; i++) {

		/* Each time a codeword becomes one longer, the current codeword
		 * is left shifted by one place.  This is part of the procedure
		 * for enumerating the canonical Huffman code.  Additionally,
		 * whenever a codeword is used, 1 is added to the current
		 * codeword.  */

		unsigned len_diff = leaves[i].path_len - cur_codeword_len;
		cur_codeword <<= len_diff;
		cur_codeword_len += len_diff;

		u16 sym = leaves[i].sym;
		codewords[sym] = cur_codeword;
		lens[sym] = cur_codeword_len;

		cur_codeword++;
	}
}


/* Retries @num_bytes bytes from the input byte buffer @buf and stores them at
 * the memory location pointed to by @p.  Returns true on success, and false if
 * there are not @num_bytes remaining in the buffer. */
static inline bool get_bytes(struct input_byte_buffer *buf, 
					unsigned num_bytes, void *p)
{
	if (num_bytes > buf->num_bytes_remaining)
		return false;
	memcpy(p, buf->data, num_bytes);
	return true;
}

/* Initializes an input byte buffer to use the data pointed to by @data, with
 * length @num_bytes bytes. */
static inline void init_input_byte_buffer(struct input_byte_buffer *buf, 
					void *data, unsigned num_bytes)
{
	buf->data = data;
	buf->num_bytes_remaining = num_bytes;
}


/* Initializes an output bit buffer to write its output to the memory location
 * pointer to by @data. */
static inline void init_output_bit_buffer(struct output_bit_buffer *buf,
						void *data, unsigned num_bytes)
{
	buf->bitbuf      = 0;
	buf->filled_bits = 0;
	buf->output      = data;
	buf->num_bytes_remaining = num_bytes;
}

/* Number of bits in a bitbuffer variable that stages bits before they are
 * written to the output memory location. */
#define BITBUF_CAPACITY (sizeof(bitbuf_t) * 8)

/* Number of 16-bit words in the bitbuffer variable. */
#define BITBUF_SIZE_WORDS (sizeof(bitbuf_t) / sizeof(u16))

/* Writes @num_bits bits, given by the @num_bits least significant bits of
 * @bits, to the output @buf. */
static bool ALWAYS_INLINE put_bits(struct output_bit_buffer *buf, 
					bitbuf_t bits, unsigned num_bits)
{
	wimlib_assert(num_bits <= BITBUF_CAPACITY);

	while (num_bits != 0) {

		/* It is very tricky to output the bits correctly.  The correct
		 * way is to output little-endian 2-byte words, such that the
		 * bits in the SECOND byte logically precede those in the FIRST
		 * byte.  While the byte order is little-endian, the bit order
		 * is big-endian; the first bit in a byte is the high-order one.
		 * Any multi-bit numbers are in bit-big-endian form, so the
		 * low-order bit of a multi-bit number is the LAST bit to be
		 * output. */
		
		/* Number of bits that we can fit in the buffer. */
		unsigned bits_consumed = min(num_bits, BITBUF_CAPACITY - 
							buf->filled_bits);
		/* Shift the existing bits in the buffer to the left (to a
		 * higher order), then OR the bits into the low-order bits of
		 * the buffer.  */
		buf->filled_bits += bits_consumed;
		buf->bitbuf <<= bits_consumed;
		buf->bitbuf |= (bits >> (num_bits - bits_consumed)) & 
				((1 << bits_consumed) - 1);

		num_bits -= bits_consumed;

		if (buf->filled_bits == BITBUF_CAPACITY) {

			/* Buffer is full--- output a 2-byte word. */

			if (buf->num_bytes_remaining < sizeof(bitbuf_t))
				return false;
			*(u16*)buf->output = to_le16(buf->bitbuf);
			buf->bitbuf = 0;
			buf->filled_bits = 0;
			buf->output += 2;
			buf->num_bytes_remaining -= 2;
		}
	}
	return true;
}

/* Flushes any remaining bits in the output buffer to the output byte stream. */
static bool flush_output_buffer(struct output_bit_buffer *buf)
{
	if (buf->filled_bits == 0)
		return true;
	if (buf->num_bytes_remaining < 2)
		return false;

	buf->bitbuf <<= (BITBUF_CAPACITY - buf->filled_bits);

	*(u16*)buf->output = to_le16(buf->bitbuf);
	buf->num_bytes_remaining -= 2;
	buf->output += 2;
	buf->filled_bits = 0;
	buf->bitbuf = 0;
	return true;
}

/* This function determines the best match for the string of characters
 * beginning at every index in the block/buffer.  The "best match" for a given
 * index is defined as the longest string of characters previously occurring in
 * the block that is the same as the string of characters starting at that
 * index.  If multiple matches tie for the longest length, the best one is taken
 * to be the last one (the one closest to the string being matched) to reduce
 * the number of bits needed to represent the match offset.
 *
 * match_offset_tab is an array indexed by buffer index.  For each character in
 * the buffer of offset i, match_offset_tab[i] will give the index of beginning
 * of the longest match for the string of characters beginning at that
 * character.
 *
 * For each character in the buffer of offset i, match_len_tab[i] will give
 * the length of beginning of the longest match for the string of
 * characters beginning at that character. 
 *
 * XXX This function takes up ~98% of the compression time. It should be
 * optimized somehow, perhaps by using a hash table like in zlib.
 */
static void lz_analyze_block(const u8 block[], int block_len, 
				u16 match_offset_tab[], u16 match_len_tab[])
{
	/* Make a pass through the block to find matches of length 1.  This
	 * means that we determine, for each character in the block, the index
	 * at which that character had last previously appeared, relative to
	 * that character. */
	{
		/* chartab holds pointers to the last occurrence of each
		 * character. */
		const u8 *char_tab[256];
		ZERO_ARRAY(char_tab);

		for (int i = 0; i < block_len; i++) {
			u8 c = block[i];
			if (char_tab[c] != NULL) {
				/* Case where the character 'c' has already
				 * appeared in the block, so we can update the
				 * i-th entry of the match_offset_tab and the
				 * match_len_tab. */
				match_offset_tab[i] = &block[i] - char_tab[c];
				match_len_tab[i] = 1;
			} else {
				/* Case where the character 'c' has not appeared
				 * in the block yet. */
				match_len_tab[i] = 0;
			}
			char_tab[c] = &block[i];
		}
	}

	/* Now, make a pass through the block for each additional allowed match
	 * length, in order.  Every pass, we look for matches that are 1 longer
	 * than the matches we found previously.  The pass is done from the end
	 * of the block to the beginning, so that matches closer to the string
	 * being matched are always found before farther away matches.  */
	for (int match_len = 2; match_len <= MAX_MATCH; match_len++) {

		/* Set to true if a match of length match_len is found.  If no
		 * such match is found, there is no need to search for longer
		 * matches, since there cannot be any. */
		bool found_a_match_of_len_match_len = false;

		for (int i = block_len - match_len; i >= 0; i--) {
			 /* For each character in the block (index i),
			  * match_offset_tab will contain the index of the best
			  * match for that location of no more than match_len -
			  * 1 bytes long, relative to that location.
			  * match_len_tab will contain the length of that match.
			  *
			  * If the length given in match_len_tab for index i is
			  * not equal to the previous match length (match_len -
			  * 1), then we have already found the longest match for
			  * the string beginning at index i, so searching for a
			  * longer one is not needed.
			  */

			if (match_len_tab[i] != match_len - 1)
				continue;

			/* c is the next character in the string that must be
			 * matched.  */
			u8 c = block[i + match_len - 1];

			/* Follow the chain of matches back to the beginning of
			 * the block to find a match for the string beginning
			 * at index i that matches match_len characters, all the
			 * way up to character c.  We can stop searching for a
			 * match whenever we get to a point when we know that
			 * there cannot be any previous match longer than
			 * match_len - 1 characters (when the match_len at the
			 * cursor index is less that match_len - 1). */
			u16 cursor_idx = i - match_offset_tab[i];
			while (match_len_tab[cursor_idx] == match_len - 1) {

				if (block[cursor_idx + match_len - 1] == c) {
					match_offset_tab[i] = i - cursor_idx;
					match_len_tab[i] = match_len;
					found_a_match_of_len_match_len = true;
					break;
				}
				cursor_idx -= match_offset_tab[cursor_idx];
			}
		}

		if (!found_a_match_of_len_match_len)
			break;
	}
}

/* Returns the position slot that corresponds to a given formatted offset.  This
 * means searching the lzx_position_base array to find what slot contains a
 * position base that is less than or equal to formatted_offset, where the next
 * slot contains a position base that is greater than or equal to
 * formatted_offset. */
static unsigned get_position_slot(unsigned formatted_offset)
{
	/* Calculate position base using binary search of table; if log2 can be
	 * done in hardware, approximation might work; 
	 * trunc(log2(formatted_offset*formatted_offset)) gets either the proper
	 * position slot or the next one, except for slots 0, 1, and 39-49
	 *
	 * Slots 0-1 are handled by the R0-R1 procedures
	 *
	 * Slots 36-49 (formatted_offset >= 262144) can be found by 
	 * (formatted_offset/131072) + 34 == (formatted_offset >> 17) + 34;
	 */
	if (formatted_offset >= 262144) {
		return (formatted_offset >> 17) + 34;
	} else {
		int left, right, mid;

		left = 3;
		right = NUM_POSITION_SLOTS - 1;
		while (1) {
			mid = (left + right) >> 1;
			if ((lzx_position_base[mid] <= formatted_offset) &&
			    lzx_position_base[mid + 1] > formatted_offset) {
				return mid;
			}
			if (formatted_offset > lzx_position_base[mid])
				/* too low */
				left = mid + 1;
			else	/* too high */
				right = mid;
		}
	}
}

/* Constructs a match from an offset and a length, and updates the LRU queue
 * and the frequency of symbols in the main, length, and aligned offset
 * alphabets.  The return value is a 32-bit integer that, if the high bit is
 * set, contains the match length, the position slot, and the position footer
 * for the match.  If it is decided that it is not worth it to do a match, 0 is
 * returned.  */
static u32 record_match(unsigned match_offset, unsigned match_len, 
				struct lru_queue *queue, 
				struct lzx_freq_tables *freq_tabs)
{
	unsigned formatted_offset;
	unsigned position_slot;
	unsigned position_footer = 0;

	if (match_len < 3)
		return 0;

	if (match_offset == queue->R0) {
		formatted_offset = 0;
		position_slot    = 0;
	} else if (match_offset == queue->R1) {
		swap(queue->R0, queue->R1);
		formatted_offset = 1;
		position_slot    = 1;
	} else if (match_offset == queue->R2) {
		swap(queue->R0, queue->R2);
		formatted_offset = 2;
		position_slot    = 2;
	} else {
		/* Not a repeated offset. */

		formatted_offset = match_offset + MIN_MATCH;

		/* reject matches where extra_bits will likely be bigger
		 * than just outputting literals.  The numbers are
		 * basically derived through guessing and trial and
		 * error */
		if ((formatted_offset >= 64 && match_len < 4) ||
		    (formatted_offset >= 2048 && match_len < 5) ||
		    (formatted_offset >= 65536 && match_len < 6)) {
			return 0;
		}

		queue->R2 = queue->R1;
		queue->R1 = queue->R0;
		queue->R0 = match_offset;

		position_slot = get_position_slot(formatted_offset);

		/* Just the extra bits of the formatted offset. */
		position_footer = ((1UL << lzx_extra_bits[position_slot]) - 1) &
								formatted_offset;
	}

	/* (match length - 2) = 8 bits */
	/* position_slot = 6 bits */
	/* position_footer = 17 bits */
	/* total = 31 bits */
	/* plus one to say whether it's a literal or not */

	u32 match = 0x80000000 | /* bit 31 in intelligent bit ordering */
		(position_slot << 25) | /* bits 30-25 */
		(position_footer << 8) | /* bits 8-24 */
		(match_len - MIN_MATCH); /* bits 0-7 */

	unsigned len_header;
	unsigned len_pos_header;

	/* Update the frequency for the main tree, the length tree (only if a
	 * length symbol is to be output), and the aligned tree (only if an
	 * aligned symbol is to be output.) */
	if (match_len < (NUM_PRIMARY_LENS + MIN_MATCH)) {
		len_header = match_len - MIN_MATCH;
	} else {
		unsigned len_footer;
		len_header = NUM_PRIMARY_LENS;
		len_footer = match_len - (NUM_PRIMARY_LENS + MIN_MATCH);
		freq_tabs->len_freq_table[len_footer]++;
	}
	len_pos_header = (position_slot << 3) | len_header;
	wimlib_assert(len_pos_header < MAINTREE_NUM_SYMBOLS - NUM_CHARS);
	freq_tabs->main_freq_table[len_pos_header + NUM_CHARS]++;

	if (lzx_extra_bits[position_slot] >= 3)
		freq_tabs->aligned_freq_table[position_footer & 7]++;

	return match;
}

/* Records a literal value, as opposed to a match.  The frequency of
 * literal_char is incremented in the table that records the frequencies of
 * symbols in the main alphabet.  Return value is simply the literal character.
 * */
static inline u32 record_literal(u8 literal_char, 
				struct lzx_freq_tables *freq_tabs)
{
	freq_tabs->main_freq_table[literal_char]++;
	return (u32)literal_char;
}

/* 
 * Writes a compressed literal match to the output.
 *
 * @out:         The output bitstream.
 * @block_type:  The type of the block (BLOCKTYPE_ALIGNED or BLOCKTYPE_VERBATIM)
 * @match:   	 The match, encoded as a 32-bit number.
 * @codes:	Pointer to a structure that contains the codewords for the
 * 			main, length, and aligned offset Huffman codes.
 */
static bool write_match(struct output_bit_buffer *out, int block_type,
				u32 match, const struct lzx_codes *codes)
{
	/* low 8 bits are the match length minus 2 */
	unsigned match_len_minus_2 = match & 0xff;
	/* Next 17 bits are the position footer */
	unsigned position_footer = (match >> 8) & 0x1ffff;	/* 17 bits */
	/* Next 6 bits are the position slot. */
	unsigned position_slot = (match >> 25) & 0x3f;	/* 6 bits */

	unsigned len_header;
	unsigned len_footer;

	/* If the match length is less than MIN_MATCH (= 2) +
	 * NUM_PRIMARY_LENS (= 7), the length header contains
	 * the match length minus MIN_MATCH, and there is no
	 * length footer.
	 *
	 * Otherwise, the length header contains
	 * NUM_PRIMARY_LENS, and the length footer contains
	 * the match length minus NUM_PRIMARY_LENS minus
	 * MIN_MATCH. */
	if (match_len_minus_2 < NUM_PRIMARY_LENS) {
		len_header = match_len_minus_2;
		/* No length footer-- mark it with a special
		 * value. */
		len_footer = (unsigned)(-1);
	} else {
		len_header = NUM_PRIMARY_LENS;
		len_footer = match_len_minus_2 - NUM_PRIMARY_LENS;
	}

	/* Combine the position slot with the length header into
	 * a single symbol that will be encoded with the main
	 * tree. */
	unsigned len_pos_header = (position_slot << 3) | len_header;

	/* The actual main symbol is offset by NUM_CHARS because
	 * values under NUM_CHARS are used to indicate a literal
	 * byte rather than a match. */
	unsigned main_symbol = len_pos_header + NUM_CHARS;

	/* Output main symbol. */
	if (!put_bits(out, codes->main_codewords[main_symbol], 
				codes->main_lens[main_symbol]))
		return false;

	/* If there is a length footer, output it using the
	 * length Huffman code. */
	if (len_footer != (unsigned)(-1)) {
		if (!put_bits(out, codes->len_codewords[len_footer],
				codes->len_lens[len_footer]))
			return false;
	}

	wimlib_assert(position_slot < NUM_POSITION_SLOTS);
	unsigned num_extra_bits = lzx_extra_bits[position_slot];

	/* For aligned offset blocks with at least 3 extra bits, output the
	 * verbatim bits literally, then the aligned bits encoded using the
	 * aligned offset tree.  Otherwise, only the verbatim bits need to be
	 * output. */
	if ((block_type == BLOCKTYPE_ALIGNED) && (num_extra_bits >= 3))
	{
		unsigned verbatim_bits = position_footer >> 3;
		if (!put_bits(out, verbatim_bits, num_extra_bits - 3))
			return false;

		unsigned aligned_bits = (position_footer & 7);
		if (!put_bits(out, codes->aligned_codewords[aligned_bits],
				codes->aligned_lens[aligned_bits]))
			return false;
	} else {
		/* verbatim bits is the same as the position
		 * footer, in this case. */
		if (!put_bits(out, position_footer, num_extra_bits))
			return false;
	}
	return true;
}

/* 
 * Writes all compressed literals in a block, both matches and literal bytes, to
 * the output bitstream.
 *
 * @out:         The output bitstream.
 * @block_type:  The type of the block (BLOCKTYPE_ALIGNED or BLOCKTYPE_VERBATIM)
 * @matches[]:   The array of matches that will be output.  It has length
 * 			of @num_compressed_literals.
 * @num_compressed_literals:  Number of compressed literals to be output.
 * @codes:	Pointer to a structure that contains the codewords for the
 * 			main, length, and aligned offset Huffman codes.
 *
 * Returns true on success, and false if the output ran out of space.
 */
static bool write_compressed_literals(struct output_bit_buffer *out, 
					int block_type,
			 		const u32 matches[], 
					unsigned  num_compressed_literals,
					const struct lzx_codes *codes)
{
	for (unsigned i = 0; i < num_compressed_literals; i++) {

		u32 match = matches[i];
		/* High bit of the match indicates whether the match is an
		 * actual match (1) or a literal uncompressed byte (0) */
		if (match & 0x80000000) {
			/* match */
			if (!write_match(out, block_type, match, codes))
				return false;
		} else {
			/* literal byte */
			wimlib_assert(match < NUM_CHARS);
			if (!put_bits(out, codes->main_codewords[match],
					codes->main_lens[match]))
				return false;
		}
	}
	return true;
}

/* 
 * Writes a compressed Huffman tree to the output, preceded by the pretree for
 * it.
 *
 * The Huffman tree is represented in the output as a series of path lengths
 * from which the canonical Huffman code can be reconstructed.  The path lengths
 * themselves are compressed using a separate Huffman code, the pretree, which
 * consists of PRETREE_NUM_SYMBOLS (= 20) symbols that cover all possible code
 * lengths, plus extra codes for repeated lengths.  The path lengths of the
 * pretree precede the path lengths of the larger code and are uncompressed,
 * consisting of 20 entries of 4 bits each.
 *
 * @out:	The bitstream for the compressed output.
 * @lens:	The code lengths for the Huffman tree, indexed by symbol.
 * @num_symbols:	The number of symbols in the code.
 */
static int write_compressed_tree(struct output_bit_buffer *out, 
				const u8 lens[], 
				unsigned num_symbols)
{
	/* Frequencies of the length symbols, including the RLE symbols (NOT the
	 * actual lengths themselves). */
	unsigned pretree_freqs[PRETREE_NUM_SYMBOLS];
	ZERO_ARRAY(pretree_freqs);

	/* Since the code word lengths use a form of RLE encoding, the goal here
	 * is to find each run of identical lengths when going through them in
	 * symbol order (including runs of length 1).  For each run, as many
	 * lengths are encoded using RLE as possible, and the rest are output
	 * literally.
	 *
	 * output_syms[] will be filled in with the length symbols that will be
	 * output, including RLE codes, not yet encoded using the pre-tree.
	 *
	 * cur_run_len keeps track of how many code word lengths are in the
	 * current run of identical lengths.
	 */
	u8 output_syms[num_symbols * 2];
	unsigned output_syms_idx = 0;
	unsigned cur_run_len = 1;
	for (unsigned i = 1; i <= num_symbols; i++) {

		if (i != num_symbols && lens[i] == lens[i - 1]) {
			/* Still in a run--- keep going. */
			cur_run_len++;
			continue;
		}

		/* Run ended! Check if it is a run of zeroes or a run of
		 * nonzeroes. */

		/* The symbol that was repeated in the run--- not to be confused
		 * with the length *of* the run (cur_run_len) */
		unsigned len_in_run = lens[i - 1];

		if (len_in_run == 0) {
			/* A run of 0's.  Encode it in as few length
			 * codes as we can. */

			/* The magic length 18 indicates a run of 20 + n zeroes,
			 * where n is an uncompressed literal 5-bit integer that
			 * follows the magic length. */
			while (cur_run_len >= 20) {

				unsigned additional_bits = 
						min(cur_run_len - 20, 0x1f);
				pretree_freqs[18]++;
				output_syms[output_syms_idx++] = 18;
				output_syms[output_syms_idx++] = additional_bits;
				cur_run_len -= 20 + additional_bits;
			}

			/* The magic length 17 indicates a run of 4 + n zeroes,
			 * where n is an uncompressed literal 4-bit integer that
			 * follows the magic length. */
			while (cur_run_len >= 4) {
				unsigned additional_bits = 
						min(cur_run_len - 4, 0xf);
				pretree_freqs[17]++;
				output_syms[output_syms_idx++] = 17;
				output_syms[output_syms_idx++] = additional_bits;
				cur_run_len -= 4 + additional_bits;
			}

		} else {

			/* A run of nonzero lengths. */

			/* The magic length 19 indicates a run of 4 + n
			 * nonzeroes, where n is a literal bit that follows the
			 * magic length, and where the value of the lengths in
			 * the run is given by an extra length symbol, encoded
			 * with the pretree, that follows the literal bit.
			 *
			 * The extra length symbol is encoded as a difference
			 * from the length of the codeword for the first symbol
			 * in the run in the previous tree.
			 * */
			while (cur_run_len >= 4) {
				unsigned additional_bits = (cur_run_len > 4);
				char delta = -(char)len_in_run;
				if (delta < 0)
					delta += 17;
				pretree_freqs[19]++;
				pretree_freqs[delta]++;
				output_syms[output_syms_idx++] = 19;
				output_syms[output_syms_idx++] = additional_bits;
				output_syms[output_syms_idx++] = delta;
				cur_run_len -= 4 + additional_bits;
			}
		}

		/* Any remaining lengths in the run are outputted without RLE,
		 * as a difference from the length of that codeword in the
		 * previous tree. */
		while (cur_run_len--) {
			char delta = -(char)len_in_run;
			if (delta < 0)
				delta += 17;

			pretree_freqs[delta]++;
			output_syms[output_syms_idx++] = delta;
		}

		cur_run_len = 1;
	}

	wimlib_assert(output_syms_idx < ARRAY_LEN(output_syms));

	/* Build the pretree from the frequencies of the length symbols. */

	u8 pretree_lens[PRETREE_NUM_SYMBOLS];
	u16 pretree_codewords[PRETREE_NUM_SYMBOLS];
	make_canonical_huffman_code(PRETREE_NUM_SYMBOLS, MAX_CODE_LEN, 
					pretree_freqs, pretree_lens, 
					pretree_codewords);

	/* Write the lengths of the pretree codes to the output. */
	for (int i = 0; i < PRETREE_NUM_SYMBOLS; i++)
		put_bits(out, pretree_lens[i], PRETREE_ELEMENT_SIZE);

	/* Write the length symbols, encoded with the pretree, to the output. */

	unsigned i = 0;
	while (i < output_syms_idx) {
		u8 pretree_sym = output_syms[i++];

		put_bits(out, pretree_codewords[pretree_sym], 
						pretree_lens[pretree_sym]);
		switch (pretree_sym) {
		case 17:
			put_bits(out, output_syms[i++], 4);
			break;
		case 18:
			put_bits(out, output_syms[i++], 5);
			break;
		case 19:
			put_bits(out, output_syms[i++], 1);
			put_bits(out, pretree_codewords[output_syms[i]],
					pretree_lens[output_syms[i]]);
			i++;
			break;
		default:
			break;
		}
	}
	return true;
}

/* Builds the canonical Huffman code for the main tree, the length tree, and the
 * aligned offset tree. */
static void make_huffman_codes(const struct lzx_freq_tables *freq_tabs,
				struct lzx_codes *codes)
{
	make_canonical_huffman_code(MAINTREE_NUM_SYMBOLS, MAX_CODE_LEN,
					freq_tabs->main_freq_table, 
					codes->main_lens,
					codes->main_codewords);

	make_canonical_huffman_code(LENTREE_NUM_SYMBOLS, MAX_CODE_LEN,
					freq_tabs->len_freq_table, 
					codes->len_lens,
					codes->len_codewords);

	make_canonical_huffman_code(ALIGNEDTREE_NUM_SYMBOLS, 8,
					freq_tabs->aligned_freq_table, 
					codes->aligned_lens,
					codes->aligned_codewords);
}

/* Do the 'E8' preprocessing, where the targets of x86 CALL instructions were
 * changed from relative offsets to absolute offsets.  This type of
 * preprocessing can be used on any binary data even if it is not actually
 * machine code.  It seems to always be used in WIM files, even though there is
 * no bit to indicate that it actually is used, unlike in the LZX compressed
 * format as used in other file formats such as the cabinet format, where a bit
 * is reserved for that purpose. */
static void do_call_insn_preprocessing(u8 uncompressed_data[], 
						unsigned uncompressed_data_len)
{
	int i = 0;
	int file_size = MAGIC_FILESIZE;

	/* Not enabled in the last 6 bytes, which means the 5-byte call
	 * instruction cannot start in the last *10* bytes. */
	while (i < uncompressed_data_len - 10) { 
		if (uncompressed_data[i] != 0xe8) {
			i++;
			continue;
		}
		int32_t rel_offset = to_le32(*(int32_t*)(uncompressed_data + i + 1));

		if (rel_offset >= -i && rel_offset < file_size) {
			int32_t abs_offset;
			if (rel_offset < file_size - i) {
				/* "good translation" */
				abs_offset = rel_offset + i;
			} else {
				/* "compensating translation" */
				abs_offset = rel_offset - file_size;
			}
			*(int32_t*)(uncompressed_data + i + 1) = to_le32(abs_offset);
		}
		i += 5;
	}
}

/* 
 * Performs LZX compression on a block of data.
 *
 * @__compressed_data:	Pointer to the data to be compressed.
 * @uncompressed_len:	Length, in bytes, of the data to be compressed.
 * @compressed_data:	Pointer to a location at least (@uncompressed_len - 1)
 * 				bytes long into which the compressed data may be
 * 				written.
 * @compressed_len_ret:	A pointer to an unsigned int into which the length of
 * 				the compressed data may be returned.
 *
 * Returns true if compression was successfully performed.  In that case
 * @compressed_data and @compressed_len_ret will contain the compressed data and
 * its length.  A return value of false means that compressing the data did not
 * reduce its size, and @compressed_data will not contain the full compressed
 * data. 
 */
bool lzx_compress(const void *__uncompressed_data, 
			unsigned uncompressed_len,
			void *compressed_data, unsigned *compressed_len_ret)
{
	if (uncompressed_len < 100)
		return false;

	LZX_DEBUG("uncompressed_len = %u\n", uncompressed_len);

	struct input_byte_buffer input_buf;
	struct output_bit_buffer output_buf;

	u8 uncompressed_data[uncompressed_len];

	u16 match_offset_tab[uncompressed_len];
	u16 match_len_tab[uncompressed_len];
	u32 matches[uncompressed_len];

	struct lzx_freq_tables freq_tabs;
	struct lzx_codes codes;
	struct lru_queue queue = {.R0 = 1, .R1 = 1, .R2 = 1};

	memcpy(uncompressed_data, __uncompressed_data, uncompressed_len);

	memset(&freq_tabs, 0, sizeof(freq_tabs));


	/* Initialize the input and output buffers. */
	init_input_byte_buffer(&input_buf, uncompressed_data, 
						uncompressed_len);
	init_output_bit_buffer(&output_buf, compressed_data, 
						uncompressed_len - 1);

	LZX_DEBUG("Doing call instruction preprocessing.\n");

	/* Before doing any actual compression, do the call instruction (0xe8
	 * byte) translation on the uncompressed data. */
	do_call_insn_preprocessing(uncompressed_data, uncompressed_len);

	LZX_DEBUG("Analyzing block for matches.\n");

	/* Analyze the block for matches.  This determines the best match for
	 * every character in the input.  Not all of these matches are
	 * actually used. */
	lz_analyze_block(uncompressed_data, uncompressed_len, match_offset_tab,
								match_len_tab);

	/* Now, figure out what matches will actually be used.  Some will
	 * obviously go unused because matches cover multiple characters, but
	 * also there can be literal characters output. In the process build the
	 * frequency tables for the main, length, and aligned trees. */

	LZX_DEBUG("Recording matches.\n");

	unsigned cur_match_pos = 0;
	unsigned cur_input_pos = 0;
	while (cur_input_pos < uncompressed_len) {
		u32 match = record_match(match_offset_tab[cur_input_pos],
						match_len_tab[cur_input_pos],
							&queue, &freq_tabs);
		if (match == 0) {
			match = record_literal(uncompressed_data[cur_input_pos], 
						&freq_tabs);
			cur_input_pos++;
		} else {
			wimlib_assert(match & 0x80000000);
			cur_input_pos += match_len_tab[cur_input_pos];
		}

		matches[cur_match_pos++] = match;
	}

	unsigned num_matches = cur_match_pos;

	LZX_DEBUG("using %u matches\n", num_matches);

	wimlib_assert(cur_input_pos == uncompressed_len);

	LZX_DEBUG("Making Huffman codes.\n");

	make_huffman_codes(&freq_tabs, &codes);

	LZX_DEBUG("Writing header.\n");

	/* The first three bits tell us what kind of block it is, and are one
	 * of the BLOCKTYPE_* values.  */
	put_bits(&output_buf, BLOCKTYPE_ALIGNED, 3);

	/* The next bit indicates whether the block size is the default (32768),
	 * indicated by a 1 bit, or whether the block size is given by the next
	 * 16 bits, indicated by a 0 bit. */
	if (uncompressed_len == 32768) {
		put_bits(&output_buf, 1, 1);
	} else {
		put_bits(&output_buf, 0, 1);
		put_bits(&output_buf, uncompressed_len, 16);
	}

	LZX_DEBUG("Writing aligned tree.\n");

	/* Write out the aligned offset tree. Note that M$ lies and says that
	 * the aligned offset tree comes after the length tree, but that is
	 * wrong; it actually is before the main tree.  */
	for (int i = 0; i < ALIGNEDTREE_NUM_SYMBOLS; i++)
		put_bits(&output_buf, codes.aligned_lens[i], ALIGNEDTREE_ELEMENT_SIZE);

	LZX_DEBUG("Writing main tree.\n");

	/* Write the pre-tree and lengths for the first NUM_CHARS symbols in the
	 * main tree. */
	if (!write_compressed_tree(&output_buf, codes.main_lens, NUM_CHARS))
		return false;

	/* Write the pre-tree and symbols for the rest of the main tree. */
	if (!write_compressed_tree(&output_buf, codes.main_lens + NUM_CHARS, 
					MAINTREE_NUM_SYMBOLS - NUM_CHARS))
		return false;

	LZX_DEBUG("Writing length tree.\n");

	/* Write the pre-tree and symbols for the length tree. */
	if (!write_compressed_tree(&output_buf, codes.len_lens, LENTREE_NUM_SYMBOLS))
		return false;

	LZX_DEBUG("Writing compressed literals.\n");

	/* Write the compressed literals. */
	if (!write_compressed_literals(&output_buf, BLOCKTYPE_ALIGNED, matches, 
					num_matches, &codes))
		return false;

	LZX_DEBUG("Flushing output buffer.\n");

	if (!flush_output_buffer(&output_buf))
		return false;

	unsigned compressed_len = output_buf.output - (u8*)compressed_data;

	LZX_DEBUG("Compressed %u => %u bytes\n",
			uncompressed_len, compressed_len);

	*compressed_len_ret = compressed_len;

#ifdef ENABLE_LZX_VERIFY
	/* Verify that we really get the same thing back when decompressing. */
	u8 buf[uncompressed_len];
	if (!lzx_decompress(compressed_data, compressed_len, buf, uncompressed_len)) {
		wimlib_error("ERROR: Failed to decompress data we compressed!\n");
		abort();
	}
	if (memcmp(buf, __uncompressed_data, uncompressed_len) != 0) {
		wimlib_error("ERROR: Data we compressed didn't decompress to the "
				"original data!\n");
		abort();
	}
	LZX_DEBUG("Compression verified to be correct.\n");
#endif

	return true;
}
