/*
 * lzx-decomp.c
 *
 * Routines for LZX decompression.  The LZX format has many similarities to the
 * DEFLATE format used in zlib and gzip, but it's not quite the same.
 *
 *  source:     modified lzx.c from cabextract v0.5                        
 *  notes:      This file has been modified from code taken from cabextract
 *  			v0.5, which was, itself, a modified version of the 
 *  			lzx decompression code from unlzx.                                               
 *
 * wimlib - Library for working with WIM files 
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

/* 
 * This file has been customized somewhat for WIMLIB.
 *
 * Some notes on the LZX compression format as used in Windows Imaging (WIM)
 * files:
 *
 * A compressed WIM file resource consists of a table of chunk offsets followed
 * by compressed chunks.  All compressed chunks except the last decompress to
 * WIM_CHUNK_SIZE (= 32768) bytes.  This is quite similar to the cabinet (.cab)
 * file format, but they are not the same (at least based on M$'s
 * documentation).  According to the documentation, in the cabinet format, the
 * LZX block size is independent from the CFDATA blocks and may span several
 * CFDATA blocks.  However, for WIM file resources, I have seen no case of a LZX
 * block spanning multiple WIM chunks.  This is probably done to make it easier
 * to randomly access the compressed file resources.  WIMLIB in fact makes use
 * of this feature to allow semi-random access to file resources in the
 * read_resource() function.
 *
 * Usually a WIM chunk will contain only one LZX block, but on rare occasions it
 * may contain multiple LZX block. The LZX block are usually the aligned block
 * type or verbatim block type, but can (very rarely) be the uncompressed block
 * type.  The size of a LZX block is specified by 1 or 17 bits following the 3
 * bits that specify the block type.  A '1' means to use the default block size
 * (equal to 32768), while a '0' means that the block size is given by the next
 * 16 bits.
 *
 * The cabinet format, as documented, allows for the possibility that a CFDATA
 * chunk is up to 6144 bytes larger than the uncompressed data.  In the WIM
 * format, however, it appears that every chunk that would be 32768 bytes or
 * more when compressed, is actually stored uncompressed.  This is not
 * documented by M$.
 *
 * The 'e8' preprocessing step that changes x86 call instructions to use
 * absolute offsets instead of relative offsets relies on a filesize parameter.
 * There is no such parameter for this in the WIM files (even though the size of
 * the file resource could be used for this purpose), and instead a magic file
 * size of 12000000 is used.  The 'e8' preprocessing is always done, and there
 * is no bit to indicate whether it is done or not.
 *
 */

#include "util.h"
#include "endianness.h"
#include "lzx.h"
#include <string.h>


/* Huffman decoding tables and maps from symbols to code lengths. */
struct lzx_tables {

	u16 maintree_decode_table[(1 << MAINTREE_TABLEBITS) + 
					(MAINTREE_NUM_SYMBOLS * 2)];
	u8 maintree_lens[MAINTREE_NUM_SYMBOLS];


	u16 lentree_decode_table[(1 << LENTREE_TABLEBITS) + 
					(LENTREE_NUM_SYMBOLS * 2)];
	u8 lentree_lens[LENTREE_NUM_SYMBOLS];


	u16 alignedtree_decode_table[(1 << ALIGNEDTREE_TABLEBITS) + 
					(ALIGNEDTREE_NUM_SYMBOLS * 2)];
	u8 alignedtree_lens[ALIGNEDTREE_NUM_SYMBOLS];
};


/* 
 * Some more notes about errors in Microsoft's documentation:
 *
 * Microsoft's LZX document and their implementation of the
 * com.ms.util.cab Java package do not concur.
 *
 * In the LZX document, there is a table showing the correlation between
 * window size and the number of position slots. It states that the 1MB
 * window = 40 slots and the 2MB window = 42 slots. In the implementation,
 * 1MB = 42 slots, 2MB = 50 slots. The actual calculation is 'find the
 * first slot whose position base is equal to or more than the required
 * window size'. This would explain why other tables in the document refer
 * to 50 slots rather than 42.
 *
 * The constant NUM_PRIMARY_LENS used in the decompression pseudocode
 * is not defined in the specification.
 *
 * The LZX document states that aligned offset blocks have their aligned
 * offset huffman tree AFTER the main and length trees. The implementation
 * suggests that the aligned offset tree is BEFORE the main and length
 * trees.
 *
 * The LZX document decoding algorithm states that, in an aligned offset
 * block, if an extra_bits value is 1, 2 or 3, then that number of bits
 * should be read and the result added to the match offset. This is
 * correct for 1 and 2, but not 3, where just a huffman symbol (using the
 * aligned tree) should be read.
 *
 * Regarding the E8 preprocessing, the LZX document states 'No translation
 * may be performed on the last 6 bytes of the input block'. This is
 * correct.  However, the pseudocode provided checks for the *E8 leader*
 * up to the last 6 bytes. If the leader appears between -10 and -7 bytes
 * from the end, this would cause the next four bytes to be modified, at
 * least one of which would be in the last 6 bytes, which is not allowed
 * according to the spec.
 *
 * The specification states that the huffman trees must always contain at
 * least one element. However, many CAB files contain blocks where the
 * length tree is completely empty (because there are no matches), and
 * this is expected to succeed.
 */

/* Must be at least 32 bits. */
typedef unsigned long bitbuf_t;

/* Structure to provide a bitstream. */
struct bit_buffer {

	/* A variable of length at least 32 bits that is used to hold bits that
	 * have been read from the stream.  The bits are ordered from high-order
	 * to low-order; the next bit is always the high-order bit. */
	bitbuf_t   bitbuf;

	/* Number of bits in @bitbuf that are valid. */
	int        bitsleft;

	/* Pointer to the next word to be retrieved from the input. */
	const u16 *data;

	/* Number of words of data that are left. */
	int        data_bytes_left;
};

/* Initializes a bitstream to receive its input from @data. */
static inline void init_bitstream(struct bit_buffer *buf, const void *data,
					int num_data_bytes)
{
	buf->bitbuf          = 0;
	buf->bitsleft        = 0;
	buf->data            = data;
	buf->data_bytes_left = num_data_bytes;
}

/* Ensures that the bit buffer contains @num_bits bits. */
static inline bool ensure_bits(struct bit_buffer *buf, int num_bits)
{
	while (buf->bitsleft < num_bits) {
		if (buf->data_bytes_left < 2)
			return false;
		unsigned shift = sizeof(bitbuf_t) * 8 - 16 - buf->bitsleft;
		buf->bitbuf |= (bitbuf_t)to_le16(*buf->data) << shift;
		buf->data++;
		buf->bitsleft += 16;
		buf->data_bytes_left -= 2;
	}
	return true;
}

/* Returns the next @num_bits bits in the bit buffer.  It must contain at least
 * @num_bits bits to call this function. */
static inline int peek_bits(const struct bit_buffer *buf, int num_bits)
{
	return buf->bitbuf >> (sizeof(bitbuf_t) * 8 - num_bits);
}

/* Removes @num_bits bits from the bit buffer.  It must contain at least
 * @num_bits bits to call this function. */
static inline void remove_bits(struct bit_buffer *buf, int num_bits)
{
	buf->bitbuf <<= num_bits;
	buf->bitsleft -= num_bits;
}

/* Reads and returns @num_bits bits from the input bitstream. */
static inline bool read_bits(struct bit_buffer *buf, int num_bits, int *n)
{
	if (!ensure_bits(buf, num_bits)) {
		wimlib_error("Input buffer exhausted!\n");
		return false;
	}
	*n = peek_bits(buf, num_bits);
	remove_bits(buf, num_bits);
	return true;
}

/* Reads @num_bits bits from the bit buffer without checking to see if that many
 * bits are in the buffer or not. */
static inline int read_bits_nocheck(struct bit_buffer *buf, int num_bits)
{
	int n = peek_bits(buf, num_bits);
	remove_bits(buf, num_bits);
	return n;
}

/* Returns the number of bits remaining in the bit buffer variable (not data
 * words). */
static inline int get_bitsleft(const struct bit_buffer *buf)
{
	return buf->bitsleft;
}

/* Removes the bits that have been read into the bit buffer variable. */
static inline void flush_buffer(struct bit_buffer *buf)
{
	remove_bits(buf, buf->bitsleft);
	buf->bitsleft = 0;
	buf->bitbuf   = 0;
}

/* Aligns the bitstream on a 16-bit boundary.
 *
 * Note: M$'s idea of "alignment" means that for some reason, a 16-bit word
 * should be skipped over if the buffer happens to already be aligned on such a
 * boundary.  This only applies for realigning the stream after the blocktype and
 * length fields of an uncompressed block, however; it does not apply when
 * realigning the stream after the end of the uncompressed block.
 */
static bool align_bitstream_on_word_boundary(struct bit_buffer *buf,
						bool skip_word_if_aligned)
{
	if (buf->bitsleft % 16 != 0) {
		remove_bits(buf, buf->bitsleft % 16);
	} else if (skip_word_if_aligned) {
		if (buf->bitsleft == 0) {
			if (!ensure_bits(buf, 16)) {
				wimlib_error("Unexpected end of input when "
						"aligning bitstream!\n");
				return false;
			}
		}
		remove_bits(buf, 16);
	}
	return true;
}

/* Reads @n bytes from the bitstream @buf into the location pointed to by @dest.
 * The bitstream must be 16-bit aligned. */
static bool read_bytes(struct bit_buffer *buf, size_t n, void *dest)
{
	/* Precondition:  The bitstream is 16-byte aligned. */
	wimlib_assert(buf->bitsleft % 16 == 0);

	u8 *p = dest;

	/* Get the bytes currently in the buffer variable. */
	while (buf->bitsleft != 0) {
		if (n-- == 0)
			return true;
		*p++ = peek_bits(buf, 8);
		remove_bits(buf, 8);
	}

	/* Get the rest directly from the pointer to the data.  Of course, it's
	 * necessary to check there are really n bytes available. */
	if (n > buf->data_bytes_left) {
		wimlib_error("Unexpected end of input when "
				"reading %zu bytes from bitstream "
				"(only have %u bytes left)\n", n,
				buf->data_bytes_left);
		return false;
	}
	memcpy(p, buf->data, n);
	buf->data += (n >> 1);
	buf->data_bytes_left -= n;

	/* It's possible to copy an odd number of bytes and leave the buffer in
	 * an inconsistent state. Fix it by reading the next byte, if it is
	 * there. */
	if ((n & 1) && buf->data_bytes_left != 0) {
		buf->bitsleft = 8;
		buf->data_bytes_left--;
		buf->bitbuf |= (bitbuf_t)*((u8*)buf->data + 1) << 
					(sizeof(bitbuf_t) * 8 - 8);
		buf->data++;
	}
	return true;
}

/* 
 * make_decode_table(num_syms, num_bits, lengths[], decode_table[])
 *
 * Builds a fast huffman decoding table from a canonical huffman code lengths
 * table.  Based on code written by David Tritscher.
 *
 * @decode_table:	The array in which to create the fast huffman decoding
 * 				table.  It must have a length of at least
 * 				(2**num_bits) + 2 * num_syms to guarantee
 * 				that there is enough space.
 *
 * @num_syms: 	Total number of symbols in the Huffman tree.
 *
 * @num_bits:	Any symbols with a code length of num_bits or less can be
 * 			decoded in one lookup of the table.  2**num_bits
 * 			must be greater than or equal to @num_syms if there are 
 * 			any Huffman codes longer than @num_bits.
 *
 * @lengths:	An array of length @num_syms, indexable by symbol, that
 * 			gives the length of that symbol.  Because the Huffman
 * 			tree is in canonical form, it can be reconstructed by
 * 			only knowing the length of the code for each symbol.
 *
 * Returns 0 on success; returns 1 if the length values do not correspond to a
 * valid Huffman tree, or if there are codes of length greater than @num_bits
 * but 2**num_bits < num_syms.
 *
 * What exactly is the format of the fast Huffman decoding table?  The first 
 * (1 << num_bits) entries of the table are indexed by chunks of the input of
 * size @num_bits.  If the next Huffman code in the input happens to have a
 * length of exactly @num_bits, the symbol is simply read directly from the
 * decoding table.  Alternatively, if the next Huffman code has length _less
 * than_ @num_bits, the symbol is also read directly from the decode table; this
 * is possible because every entry in the table that is indexed by an integer
 * that has the shorter code as a binary prefix is filled in with the
 * appropriate symbol.  If a code has length n <= num_bits, it will have
 * 2**(num_bits - n) possible suffixes, and thus that many entries in the
 * decoding table.
 *
 * It's a bit more complicated if the next Huffman code has length of more than
 * @num_bits.  The table entry indexed by the first @num_bits of that code
 * cannot give the appropriate symbol directly, because that entry is guaranteed
 * to be referenced by the Huffman codes for multiple symbols.  And while the
 * LZX compression format does not allow codes longer than 16 bits, a table of
 * size (2 ** 16) = 65536 entries would be too slow to create.
 *
 * There are several different ways to make it possible to look up the symbols
 * for codes longer than @num_bits.  A common way is to make the entries for the
 * prefixes of length @num_bits of those entries be pointers to additional
 * decoding tables that are indexed by some number of additional bits of the
 * code symbol.  The technique used here is a bit simpler, however.  We just
 * store the needed subtrees of the Huffman tree in the decoding table after the
 * lookup entries, beginning at index (2**num_bits).  Real pointers are
 * replaced by indices into the decoding table, and we distinguish symbol
 * entries from pointers by the fact that values less than @num_syms must be
 * symbol values.
 */
static bool HOT
make_decode_table(u16 decode_table[],  unsigned num_syms, 
			unsigned num_bits, const u8 lengths[])
{
	/* Number of entries in the decode table. */
	u32 table_num_entries = 1 << num_bits;

	/* Current position in the decode table. */
	u32 decode_table_pos = 0;

	/* Fill entries for codes short enough for a direct mapping.  Here we
	 * are taking advantage of the ordering of the codes, since they are for
	 * a canonical Huffman tree.  It must be the case that all the codes of
	 * some length @code_length, zero-extended or one-extended, numerically
	 * precede all the codes of length @code_length + 1.  Furthermore, if we
	 * have 2 symbols A and B, such that A is listed before B in the lengths
	 * array, and both symbols have the same code length, then we know that
	 * the code for A numerically precedes the code for B.
	 * */
	for (unsigned code_len = 1; code_len <= num_bits; code_len++) {

		/* Number of entries that a code of length @code_length would
		 * need.  */
		u32 code_num_entries = 1 << (num_bits - code_len);


		/* For each symbol of length @code_len, fill in its entries in
		 * the decode table. */
		for (unsigned sym = 0; sym < num_syms; sym++) {

			if (lengths[sym] != code_len)
				continue;


			/* Check for table overrun.  This can only happen if the
			 * given lengths do not correspond to a valid Huffman
			 * tree.  */
			if (decode_table_pos >= table_num_entries) {
				wimlib_error("Huffman decoding table overrun: "
						"pos = %u, num_entries = %u\n",
						decode_table_pos, 
						table_num_entries);
				return false;
			}

			/* Fill all possible lookups of this symbol with
			 * the symbol itself. */
			for (unsigned i = 0; i < code_num_entries; i++)
				decode_table[decode_table_pos + i] = sym;

			/* Increment the position in the decode table by
			 * the number of entries that were just filled
			 * in. */
			decode_table_pos += code_num_entries;
		}
	}

	/* If all entries of the decode table have been filled in, there are no
	 * codes longer than num_bits, so we are done filling in the decode
	 * table. */
	if (decode_table_pos == table_num_entries)
		return true;

	/* Otherwise, fill in the remaining entries, which correspond to codes longer
	 * than @num_bits. */


	/* First, zero out the rest of the entries; this is necessary so
	 * that the entries appear as "unallocated" in the next part.  */
	for (unsigned i = decode_table_pos; i < table_num_entries; i++)
		decode_table[i] = 0;

	/* Assert that 2**num_bits is at least num_syms.  If this wasn't the
	 * case, we wouldn't be able to distinguish pointer entries from symbol
	 * entries. */
	wimlib_assert((1 << num_bits) >= num_syms);


	/* The current Huffman code.  */
	unsigned current_code = decode_table_pos;

	/* The tree nodes are allocated starting at
	 * decode_table[table_num_entries].  Remember that the full size of the
	 * table, including the extra space for the tree nodes, is actually
	 * 2**num_bits + 2 * num_syms slots, while table_num_entries is only
	 * 2**num_bits. */
	unsigned next_free_tree_slot = table_num_entries;

	/* Go through every codeword of length greater than @num_bits.  Note:
	 * the LZX format guarantees that the codeword length can be at most 16
	 * bits. */
	for (unsigned code_len = num_bits + 1; code_len <= MAX_CODE_LEN; 
							code_len++) 
	{
		current_code <<= 1;
		for (unsigned sym = 0; sym < num_syms; sym++) {
			if (lengths[sym] != code_len)
				continue;


			/* i is the index of the current node; find it from the
			 * prefix of the current Huffman code. */
			unsigned i = current_code >> (code_len - num_bits);

			if (i >= (1 << num_bits)) {
				wimlib_error("Invalid canonical Huffman code!\n");
				return false;
			}

			/* Go through each bit of the current Huffman code
			 * beyond the prefix of length num_bits and walk the
			 * tree, "allocating" slots that have not yet been
			 * allocated. */
			for (int bit_num = num_bits + 1; bit_num <= code_len; bit_num++) {

				/* If the current tree node points to nowhere
				 * but we need to follow it, allocate a new node
				 * for it to point to. */
				if (decode_table[i] == 0) {
					decode_table[i] = next_free_tree_slot;
					decode_table[next_free_tree_slot++] = 0;
					decode_table[next_free_tree_slot++] = 0;
				}

				i = decode_table[i];

				/* Is the next bit 0 or 1? If 0, go left;
				 * otherwise, go right (by incrementing i by 1) */
				int bit_pos = code_len - bit_num;

				int bit = (current_code & (1 << bit_pos)) >> 
								bit_pos;
				i += bit;
			}

			/* i is now the index of the leaf entry into which the
			 * actual symbol will go. */
			decode_table[i] = sym;

			/* Increment decode_table_pos only if the prefix of the
			 * Huffman code changes. */
			if (current_code >> (code_len - num_bits) != 
					(current_code + 1) >> (code_len - num_bits))
				decode_table_pos++;

			/* current_code is always incremented because this is
			 * how canonical Huffman codes are generated (add 1 for
			 * each code, then left shift whenever the code length
			 * increases) */
			current_code++;
		}
	}


	/* If the lengths really represented a valid Huffman tree, all
	 * @table_num_entries in the table will have been filled.  However, it
	 * is also possible that the tree is completely empty (as noted
	 * earlier) with all 0 lengths, and this is expected to succeed. */

	if (decode_table_pos != table_num_entries) {

		for (unsigned i = 0; i < num_syms; i++) {
			if (lengths[i] != 0) {
				wimlib_error("Lengths do not form a valid "
						"canonical Huffman tree "
						"(only filled %u of %u decode "
						"table slots)!\n", decode_table_pos, 
						table_num_entries);
				return false;
			}
		}
		LZX_DEBUG("Empty table!\n");
	}
	return true;
}

/* Reads a Huffman-encoded symbol when it is known there are less than
 * MAX_CODE_LEN bits remaining in the bitstream. */
static bool NOINLINE COLD
read_huffsym_near_end_of_input(struct bit_buffer *buf, 
			const u16 decode_table[], const u8 lengths[], 
			unsigned num_symbols, unsigned table_bits, int *n)
{
	int bitsleft = get_bitsleft(buf);
	int key_size;
	u16 sym;
	u16 key_bits;
	if (table_bits > bitsleft) {
		key_size = bitsleft;
		bitsleft = 0;
		key_bits = peek_bits(buf, key_size) << (table_bits - key_size);
	} else {
		key_size = table_bits;
		bitsleft -= table_bits;
		key_bits = peek_bits(buf, table_bits);
	}

	sym = decode_table[key_bits];
	if (sym >= num_symbols) {
		remove_bits(buf, key_size);
		do {
			if (bitsleft == 0) {
				wimlib_error("Input buffer exhausted!\n");
				return false;
			}
			key_bits = sym + peek_bits(buf, 1);
			remove_bits(buf, 1);
			bitsleft--;
		} while ((sym = decode_table[key_bits]) >= num_symbols);
	} else {
		remove_bits(buf, lengths[sym]);
	}
	*n = sym;
	return true;
}

/* 
 * Reads a Huffman-encoded symbol from a bitstream.
 *
 * This function may be called hundreds of millions of times when extracting a
 * large WIM file, and it is declared to be always inlined for improved
 * performance.  I'm not sure it could be made much faster that it is,
 * especially since there isn't enough time to make a big table that allows
 * decoding multiple symbols per lookup.  But if extracting files to a hard
 * disk, the IO will be the bottleneck anyway.
 *
 * @buf:	The input buffer from which the symbol will be read.
 * @decode_table:	The fast Huffman decoding table for the Huffman tree.
 * @lengths:		The table that gives the length of the code for each
 * 				symbol.
 * @num_symbols:	The number of symbols in the Huffman code.
 * @table_bits:		Huffman codes this length or less can be looked up 
 * 				directory in the decode_table, as the
 * 				decode_table contains 2**table_bits entries.
 */
static bool ALWAYS_INLINE
read_huffsym(struct bit_buffer *buf, const u16 decode_table[], 
		const u8 lengths[], unsigned num_symbols, 
		unsigned table_bits, int *n)
{
	u16 sym;
	u16 key_bits;
	/* In the most common case, there are at least MAX_CODE_LEN bits
	 * remaining in the buffer. */
	if (ensure_bits(buf, MAX_CODE_LEN)) {

		/* Use the next table_bits of the input as an index into the
		 * decode_table. */
		u16 key_bits = peek_bits(buf, table_bits);

		u16 sym = decode_table[key_bits];

		/* If the entry in the decode table is not a valid symbol, it is the
		 * offset of the root of its Huffman subtree. */
		if (sym >= num_symbols) {
			remove_bits(buf, table_bits);
			do {
				key_bits = sym + peek_bits(buf, 1);
				remove_bits(buf, 1);

				wimlib_assert(key_bits < num_symbols * 2 + (1 << table_bits));
			} while ((sym = decode_table[key_bits]) >= num_symbols);
		} else {
			if (lengths[sym] > table_bits) {
				LZX_DEBUG("lengths[%u] = %u, table_bits = %u\n",
						sym, lengths[sym], table_bits);
				for (unsigned i = 0; i < num_symbols; i++) {
					LZX_DEBUG("lens[%u] = %u\n", i, lengths[i]);
				}
			}
			wimlib_assert(lengths[sym] <= table_bits);
			remove_bits(buf, lengths[sym]);
		}
		*n = sym;
		return true;
	} else {
		/* Otherwise, we must be careful to use only the bits that are
		 * actually remaining.  Don't inline this part since it is very
		 * rarely used. */
		return read_huffsym_near_end_of_input(buf, decode_table, lengths,
					num_symbols, table_bits, n);
	}
}

/* 
 * Reads a Huffman-encoded symbol using the pre-tree. */
static inline bool read_huffsym_using_pretree(struct bit_buffer *buf, 
				const u16 pretree_decode_table[],
				const u8 pretree_lens[], int *n)
{
	return read_huffsym(buf, pretree_decode_table, pretree_lens, 
				PRETREE_NUM_SYMBOLS, PRETREE_TABLEBITS, n);
}

/* Reads a Huffman-encoded symbol using the main tree. */
static inline bool read_huffsym_using_maintree(struct bit_buffer *buf, 
				const struct lzx_tables *tables, int *n)
{
	return read_huffsym(buf, tables->maintree_decode_table, 
				tables->maintree_lens,
				MAINTREE_NUM_SYMBOLS,
				MAINTREE_TABLEBITS, n);
}

/* Reads a Huffman-encoded symbol using the length tree. */
static inline bool read_huffsym_using_lentree(struct bit_buffer *buf, 
				const struct lzx_tables *tables, int *n)
{
	return read_huffsym(buf, tables->lentree_decode_table, 
				tables->lentree_lens,
				LENTREE_NUM_SYMBOLS, 
				LENTREE_TABLEBITS, n);
}

/* Reads a Huffman-encoded symbol using the aligned offset tree. */
static inline bool read_huffsym_using_alignedtree(struct bit_buffer *buf, 
				const struct lzx_tables *tables, int *n)
{
	return read_huffsym(buf, tables->alignedtree_decode_table, 
				tables->alignedtree_lens,
				ALIGNEDTREE_NUM_SYMBOLS, 
				ALIGNEDTREE_TABLEBITS, n);
}

/* 
 * Reads the pretree from the input, then uses the pretree to decode @num_lens
 * code length values from the input. 
 *
 * @buf:	The bit buffer for the input.  It is positioned on the beginning
 * 			of the pretree for the code length values.
 * @lens:	An array that contains the length values from the previous time
 * 			the code lengths for this Huffman tree were read, or all
 * 			0's if this is the first time.  
 * @num_lens:	Number of length values to decode and return.
 *
 * @return:	true on success; false if the pretree is invalid, or the input
 * 		ended unexpectedly.
 */
static bool read_code_lens(struct bit_buffer *buf, u8 lens[], unsigned num_lens)
{
	/* Declare the decoding table and length table for the pretree. */
	u16 pretree_decode_table[(1 << PRETREE_TABLEBITS) + 
					(PRETREE_NUM_SYMBOLS * 2)];
	u8 pretree_lens[PRETREE_NUM_SYMBOLS];


	/* Read the code lengths of the pretree codes.  There are 20 lengths of
	 * 4 bits each. */
	for (int i = 0; i < PRETREE_NUM_SYMBOLS; i++) {
		int len;
		if (!read_bits(buf, PRETREE_ELEMENT_SIZE, &len))
			return false;
		pretree_lens[i] = len;
	}

	/* Make the decoding table for the pretree. */
	if (!make_decode_table(pretree_decode_table, PRETREE_NUM_SYMBOLS, 
				PRETREE_TABLEBITS, pretree_lens))
		return false;

	/* Pointer past the last length value that needs to be filled in. */
	u8 *lens_end = lens + num_lens;

	while (1) {

		/* Decode a symbol from the input.  If the symbol is between 0
		 * and 16, it is the difference from the old length.  If it is
		 * between 17 and 19, it is a special code that indicates that
		 * some number of the next lengths are all 0, or some number of
		 * the next lengths are all equal to the next symbol in the
		 * input. */
		int tree_code;
		if (!read_huffsym_using_pretree(buf, pretree_decode_table, 
						pretree_lens, &tree_code))
			return false;

		int num_zeroes, code, num_same;

		char value;
		switch (tree_code) {
		case 17: /* Run of 0's */
			if (!read_bits(buf, 4, &num_zeroes))
				return false;
			num_zeroes += 4;
			while (num_zeroes--) {
				*lens = 0;
				if (++lens == lens_end)
					return true;
			}
			break;
		case 18: /* Longer run of 0's */
			if (!read_bits(buf, 5, &num_zeroes))
				return false;
			num_zeroes += 20;
			while (num_zeroes--) {
				*lens = 0;
				if (++lens == lens_end)
					return true;
			}
			break;
		case 19: /* Run of identical lengths */
			if (!read_bits(buf, 1, &num_same))
				return false;
			num_same += 4;

			if (!read_huffsym_using_pretree(buf, 
						pretree_decode_table, 
						pretree_lens, &code))
				return false;
			value = (char)*lens - (char)code;
			if (value < 0)
				value += 17;
			while (num_same--) {
				*lens = value;
				if (++lens == lens_end)
					return true;
			}
			break;
		default: /* Difference from old length. */
			value = (char)*lens - (char)tree_code;
			if (value < 0)
				value += 17;
			*lens = value;
			if (++lens == lens_end)
				return true;
			break;
		}
	}
}

/* 
 * Reads the header for an LZX-compressed block.
 *
 * @buf:		The input bitstream.
 * @block_size_ret:	A pointer to an int into which the size of the block,
 * 				in bytes, will be returned.
 * @block_type_ret:	A pointer to an int into which the type of the block
 * 				(BLOCKTYPE_*) will be returned.
 * @tables:		A pointer to a lzx_tables structure in which the 
 * 				main tree, the length tree, and possibly the
 * 				aligned offset tree will be constructed.
 * @queue:	A pointer to the least-recently-used queue into which
 * 			R0, R1, and R2 will be written (only for uncompressed
 * 			blocks, which contain this information in the header)
 *
 * @return:		True on success, false on failure (e.g. invalid
 * 				Huffman tree, invalid block type)
 */
static bool read_block_header(struct bit_buffer *buf, int *block_size_ret, 
				int *block_type_ret, struct lzx_tables *tables,
				struct lru_queue *queue)
{
	if (!ensure_bits(buf, 4)) {
		wimlib_error("Input buffer overrun!\n");
		return false;
	}

	/* The first three bits tell us what kind of block it is, and are one
	 * of the BLOCKTYPE_* values.  */
	int block_type = read_bits_nocheck(buf, 3);

	/* The next bit indicates whether the block size is the default (32768),
	 * indicated by a 1 bit, or whether the block size is given by the next
	 * 16 bits, indicated by a 0 bit. */
	int s = read_bits_nocheck(buf, 1);

	int block_size;
	if (s == 1) {
		block_size = 1 << 15;
	} else {
		if (!read_bits(buf, 16, &block_size))
			return false;
		block_size = to_le16(block_size);
	}

	switch (block_type) {
	case BLOCKTYPE_ALIGNED:

		/* Read the path lengths for the elements of the aligned tree,
		 * then build it. */
		if (!ensure_bits(buf, 24)) {
			wimlib_error("Input buffer overrun!\n");
			return false;
		}

		for (int i = 0; i < ALIGNEDTREE_NUM_SYMBOLS; i++) {
			tables->alignedtree_lens[i] = read_bits_nocheck(buf, 
						ALIGNEDTREE_ELEMENT_SIZE);
		}
		
		LZX_DEBUG("Building the aligned tree.\n");
		if (!make_decode_table(	tables->alignedtree_decode_table,
					ALIGNEDTREE_NUM_SYMBOLS, 
					ALIGNEDTREE_TABLEBITS,
					tables->alignedtree_lens))
		{
			wimlib_error("Failed to make the decode table for "
					"the aligned offset tree!\n");
			return false;
		}

		/* Fall though, since the rest of the header for aligned offset
		 * blocks is the same as that for verbatim blocks */

	case BLOCKTYPE_VERBATIM:
		if (block_type == BLOCKTYPE_VERBATIM)
			LZX_DEBUG("Found verbatim block\n");

		LZX_DEBUG("Reading path lengths for main tree.\n");
		/* Read the path lengths for the first 256 elements of the main
		 * tree. */
		if (!read_code_lens(buf, tables->maintree_lens, NUM_CHARS))
		{
			wimlib_error("Failed to read the code lengths for "
					"the first 256 elements of the main "
					"tree!\n");
			return false;
		}




		/* Read the path lengths for the remaining elements of the main
		 * tree. */
		LZX_DEBUG("Reading path lengths for remaining elements of "
				"main tree (%d elements).\n",
				MAINTREE_NUM_SYMBOLS - NUM_CHARS);
		if (!read_code_lens(buf, tables->maintree_lens + NUM_CHARS,
					MAINTREE_NUM_SYMBOLS - NUM_CHARS))
		{
			wimlib_error("Failed to read the path lengths for "
					"the remaining elements of the main "
					"tree!\n");
			return false;
		}

		LZX_DEBUG("Building the Huffman decoding table for the main tree.\n");

		if (!make_decode_table(	tables->maintree_decode_table,
					MAINTREE_NUM_SYMBOLS,
					MAINTREE_TABLEBITS,
					tables->maintree_lens))
		{
			wimlib_error("Failed to make the decode table for "
					"the main tree!\n");
			return false;
		}

		LZX_DEBUG("Reading path lengths for the length tree.\n");
		if (!read_code_lens(buf, tables->lentree_lens, 
					LENTREE_NUM_SYMBOLS))
		{
			wimlib_error("Failed to read the path lengths "
					"for the length tree!\n");
			return false;
		}

		LZX_DEBUG("Building the length tree.\n");
		if (!make_decode_table(	tables->lentree_decode_table,
					LENTREE_NUM_SYMBOLS, 
					LENTREE_TABLEBITS,
					tables->lentree_lens))
		{
			wimlib_error("Failed to build the length Huffman "
					"tree!\n");
			return false;
		}

		break;

	case BLOCKTYPE_UNCOMPRESSED:
	{
		LZX_DEBUG("Found uncompressed block\n");
		int32_t R[3];
		if (!align_bitstream_on_word_boundary(buf, true))
			return false;
		if (!read_bytes(buf, sizeof(R), R))
			return false;
		array_to_le32(R, ARRAY_LEN(3));
		queue->R0 = R[0];
		queue->R1 = R[1];
		queue->R2 = R[2];
		break;
	}

	default:
		LZX_DEBUG("Found invalid block\n");
		return false;
	}
	*block_type_ret = block_type;
	*block_size_ret = block_size;
	return true;
}

/* 
 * Decodes a compressed literal match value.  It refers to some match_offset to
 * a point earlier in the window, and some match_len, for which the data is to
 * be copied to the current position in the window.
 *
 * @main_element:	The start of the match data, as decoded using the main
 * 				tree.
 * @block_type:	The type of the block (BLOCKTYPE_ALIGNED or
 *			BLOCKTYPE_VERBATIM)
 * @bytes_remaining:	The amount of uncompressed data remaining to be
 * 				uncompressed.  It is an error if the match
 * 				is longer than @bytes_remaining.
 * @window:	A pointer to the window into which the uncompressed
 * 			data is being written.
 * @window_pos:	The current position in the window.
 * @tables:	Contains the Huffman tables for the block (main,
 * 			length, and also aligned offset only for
 * 			BLOCKTYPE_ALIGNED)
 * @queue:	The least-recently used queue for match offsets.
 * @buf:	The input bitstream.
 *
 * Returns the length of the match, or -1 on error (match would exceed
 * the amount of data needing to be uncompressed, or match refers to data before
 * the window, or the input bitstream ended unexpectedly).
 */
static int HOT
decode_match(int main_element, int block_type, int bytes_remaining,
			u8 *window, int window_pos, 
			const struct lzx_tables *tables, 
			struct lru_queue *queue,
			struct bit_buffer *buf)
{
	/* The main element is offset by 256 because values under 256 indicate a
	 * literal value. */
	main_element -= NUM_CHARS;

	/* The length header consists of the lower 3 bits of the main element.
	 * The position slot is the rest of it. */
	int length_header = main_element & NUM_PRIMARY_LENS;
	int position_slot = main_element >> 3;

	/* If the length_header is less than NUM_PRIMARY_LENS (= 7), it gives
	 * the match length as the offset from MIN_MATCH.  Otherwise, the length
	 * is given by an additional symbol encoded using the length tree,
	 * offset by 9 (MIN_MATCH + NUM_PRIMARY_LENS) */
	int match_len = MIN_MATCH + length_header;
	if (length_header == NUM_PRIMARY_LENS) {
		int additional_len;
		if (!read_huffsym_using_lentree(buf, tables, &additional_len))
			return -1;
		match_len += additional_len;
	}


	int match_offset;

	/* If the position_slot is 0, 1, or 2, the match offset is retrieved
	 * from the LRU queue.  Otherwise, the match offset is not in the LRU
	 * queue. */
	switch (position_slot) {
	case 0:
		match_offset = queue->R0;
		break;
	case 1:
		match_offset = queue->R1;
		swap(queue->R0, queue->R1);
		break;
	case 2:
		/* The queue doesn't work quite the same as a real LRU queue,
		 * since using the R2 offset doesn't bump the R1 offset down to
		 * R2. */
		match_offset = queue->R2;
		swap(queue->R0, queue->R2);
		break;
	default:
	{
		/* Otherwise, the offset was not encoded as one the offsets in
		 * the queue.  Depending on the position slot, there is a
		 * certain number of extra bits that need to be read to fully
		 * decode the match offset. */

		/* Look up the number of extra bits that need to be read. */
		int num_extra_bits = lzx_extra_bits[position_slot];
		int verbatim_bits;
		int aligned_bits;

		/* For aligned blocks, if there are at least 3 extra bits, the
		 * actual number of extra bits is 3 less, and they encode a
		 * number of 64-bit words that are added to the offset; there
		 * is then an additional symbol read using the aligned tree that
		 * specifies the actual byte alignment. */
		if (block_type == BLOCKTYPE_ALIGNED && num_extra_bits >= 3) {

			/* There is an error in the LZX "specification" at this
			 * point; it indicates that a Huffman symbol is to be
			 * read only if num_extra_bits is greater than 3, but
			 * actually it is if num_extra_bits is greater than or
			 * equal to 3.  (Note that in the case with
			 * num_extra_bits == 3, the assignment to verbatim_bits
			 * will just set it to 0. ) */
			if (!read_bits(buf, num_extra_bits - 3, &verbatim_bits))
				return false;

			verbatim_bits <<= 3;

			if (!read_huffsym_using_alignedtree(buf, tables, 
								&aligned_bits))
				return false;
		} else {
			/* For non-aligned blocks, or for aligned blocks with
			 * less than 3 extra bits, the extra bits are added
			 * directly to the match offset, and the correction for
			 * the alignment is taken to be 0. */
			if (!read_bits(buf, num_extra_bits, &verbatim_bits))
				return false;

			aligned_bits = 0;
		}

		/* Calculate the match offset. */
		match_offset = lzx_position_base[position_slot] + verbatim_bits + 
							aligned_bits - 2;

		/* Update the LRU queue. */
		queue->R2 = queue->R1;
		queue->R1 = queue->R0;
		queue->R0 = match_offset;
		break;
	}
	}

	/* Verify that the match is in the bounds of the part of the window
	 * currently in use, then copy the source of the match to the current
	 * position. */

	u8 *match_dest = window + window_pos;
	u8 *match_src = match_dest - match_offset;

	if (match_len > bytes_remaining) {
		wimlib_error("Match of length %d bytes overflows uncompressed "
				"block size!\n", match_len);
		return -1;
	}

	if (match_src < window) {
		wimlib_error("Match of length %d bytes references data "
				"before window (match_offset = %d, "
				"window_pos = %d)\n", match_len,
				match_offset, window_pos);
		return -1;
	}

	for (int i = 0; i < match_len; i++)
		match_dest[i] = match_src[i];

	return match_len;
}



/* Undo the 'E8' preprocessing, where the targets of x86 CALL instructions were
 * changed from relative offsets to absolute offsets.  This type of
 * preprocessing can be used on any binary data even if it is not actually
 * machine code.  It seems to always be used in WIM files, even though there is
 * no bit to indicate that it actually is used, unlike in the LZX compressed
 * format as used in other file formats, where a bit is reserved for that
 * purpose. */
static void undo_call_insn_preprocessing(u8 uncompressed_data[], 
						unsigned uncompressed_data_len)
{
	int i = 0;
	int file_size = MAGIC_FILESIZE;

	/* Not enabled in the last 6 bytes, which means the 5-byte call
	 * instruction cannot start in the last *10* bytes. */
	while (i < uncompressed_data_len - 10) { 
		if (uncompressed_data[i] != 0xe8) {
			i++;
			continue;
		}
		int32_t abs_offset = to_le32(*(int32_t*)(uncompressed_data + i + 1));

		if (abs_offset >= -i && abs_offset < file_size) {
			int32_t rel_offset;
			if (abs_offset >= 0) {
				/* "good translation" */
				rel_offset = abs_offset - i;
			} else {
				/* "compensating translation" */
				rel_offset = abs_offset + file_size;
			}
			*(int32_t*)(uncompressed_data + i + 1) = 
						to_le32(rel_offset);
		}
		i += 5;
	}
}

/* 
 * Decompresses a compressed block of data from which the header has already
 * been read.
 *
 * @block_type:	The type of the block (BLOCKTYPE_VERBATIM or
 * 		BLOCKTYPE_ALIGNED)
 * @block_size:	The size of the block, in bytes.
 * @window:	Pointer to the decompression window.
 * @window_pos:	The current position in the window.  Will be 0 for the first
 * 			block.	
 * @tables:	The Huffman decoding tables for the block (main, length, and
 * 			aligned offset, the latter only for BLOCKTYPE_ALIGNED)
 * @queue:	The least-recently-used queue for match offsets.
 * @buf:	The input bitstream for the compressed literals.
 */
static bool decompress_compressed_block(int block_type, 
			int block_size, u8 *window, 
			int window_pos, const struct lzx_tables *tables, 
			struct lru_queue *queue, struct bit_buffer *buf)
{
	int bytes_remaining = block_size;
	int main_element;
	while (bytes_remaining > 0) {

		if (!read_huffsym_using_maintree(buf, tables, &main_element))
			return false;

		if (main_element < NUM_CHARS) {
			/* literal: 0 to NUM_CHARS - 1 */
			window[window_pos + block_size - bytes_remaining] = 
							main_element;
			bytes_remaining--;
		} else {
			int match_len = decode_match(main_element, 
						block_type, bytes_remaining, window,
						block_size + window_pos - 
							bytes_remaining,
						tables, queue, buf);
			if (match_len == -1)
				return false;

			bytes_remaining -= match_len;
		}
	}
	return true;
}

/* 
 * Decompresses a block of LZX-compressed data using a window size of 32768.
 *
 * @compressed_data:	A pointer to the compressed data.
 * @compressed_len:	The length of the compressed data, in bytes.  
 * @uncompressed_data:	A pointer to the buffer into which to write the
 * 				uncompressed data.
 * @uncompressed_len:	The length of the uncompressed data.
 *
 * True is returned on success.  False may be returned for any of the following
 * reasons:
 * 	- The compressed data was invalid at some point; i.e., there was an
 * 	  invalid block type, or incorrect length values for a Huffman tree,
 * 	  or a match that began before the window or after the current position,
 * 	  or a block size that is greater than the remaining length of
 * 	  uncompressed data.
 * 	- @compressed_len bytes of data were used, but less than
 * 	  @uncompressed_len of uncompressed output have been written.
 */
bool lzx_decompress(const void *compressed_data, int compressed_len, 
			void *uncompressed_data, int uncompressed_len)
{
	LZX_DEBUG("lzx_decompress (compressed_data = %p, compressed_len = %d, "
			"uncompressed_data = %p, uncompressed_len = %d)\n",
			compressed_data, compressed_len, uncompressed_data,
			uncompressed_len);

#ifdef ENABLE_LZX_DEBUG
	LZX_DEBUG("compressed_data = {");
	print_byte_field(compressed_data, 32);
	puts("}");
#endif

	struct lzx_tables tables;
	struct bit_buffer buf;
	struct lru_queue  queue = {.R0 = 1, .R1 = 1, .R2 = 1};

	wimlib_assert(uncompressed_len <= 32768);

	memset(tables.maintree_lens, 0, sizeof(tables.maintree_lens));
	memset(tables.lentree_lens, 0, sizeof(tables.lentree_lens));

	init_bitstream(&buf, compressed_data, compressed_len);

	int bytes_remaining = uncompressed_len;

	/* The compressed data will consist of one or more blocks.  The
	 * following loop decompresses one block, and it runs until there all
	 * the compressed data has been decompressed, so there are no more
	 * blocks.  */

	while (bytes_remaining != 0) {
		int block_size;
		int block_type;

		LZX_DEBUG("Reading block header.\n");
		if (!read_block_header(&buf, &block_size, &block_type, &tables, &queue))
			return false;

		LZX_DEBUG("block_size = %d, bytes_remaining = %d\n",
				block_size, bytes_remaining);

		if (block_size > bytes_remaining) {
			wimlib_error("Expected a block size of at most %d "
					"bytes (found %d bytes)!\n", 
					bytes_remaining, block_size);
			return false;
		}

		if (block_type == BLOCKTYPE_VERBATIM || 
				block_type == BLOCKTYPE_ALIGNED) {
			if (block_type == BLOCKTYPE_VERBATIM)
				LZX_DEBUG("BLOCKTYPE_VERBATIM\n");
			else
				LZX_DEBUG("BLOCKTYPE_ALIGNED\n");

			if (!decompress_compressed_block(block_type, 
					   block_size, uncompressed_data,
					   uncompressed_len - bytes_remaining, 
					   &tables, &queue, &buf))
				return false;
		} else if (block_type == BLOCKTYPE_UNCOMPRESSED) {
			LZX_DEBUG("BLOCKTYPE_UNCOMPRESSED\n");
			if (!read_bytes(&buf, block_size, uncompressed_data + 
					uncompressed_len - bytes_remaining))
				return false;
			if (block_size & 1)
				align_bitstream_on_word_boundary(&buf, false);
		} else {
			wimlib_error("Unrecognized block type!\n");
			return false;
		}

		bytes_remaining -= block_size;

		if (bytes_remaining != 0) {
			LZX_DEBUG("%d bytes remaining\n", bytes_remaining);
		}

	}

	undo_call_insn_preprocessing(uncompressed_data, uncompressed_len);

	return true;
}
